import json
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from apischema import ValidationError, deserialize, serialize
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import PlainTextResponse

from ...common import logger
from ..config import StarbearServerPlugin
from .permissions import make_config


class APITokensMiddleware(BaseHTTPMiddleware):
    """Authenticate users based on API tokens.

    Arguments:
        app: The application this middleware is added to.
    """

    def __init__(self, app, mapper):
        super().__init__(app)
        self.mapper = mapper

    async def dispatch(self, request, call_next):
        key = request.headers.get("X-API-KEY", None)
        if not key:
            return await call_next(request)
        user = self.mapper(key)
        if user:
            request.session["user"] = user
            return await call_next(request)
        else:
            return PlainTextResponse("Token is invalid or expired.", status_code=401)


@dataclass
class TokenData:
    name: str
    token: str
    email: str
    # expiry: Optional[datetime.datetime] = None
    # plain: bool = True


@dataclass
class APITokens(StarbearServerPlugin):
    # Configuration file in which the tokens are located
    file: Optional[Path] = None

    # Default tokens
    defaults: Optional[list[TokenData]] = None

    def cap_require(self):
        return ["session"]

    def cap_export(self):
        return ["email"]

    def get(self, key):
        if key in self.data:
            return {"email": self.data[key].email}

    def read(self):
        return self.config.read()

    def write(self, new_tokens, dry=False):
        deserialize(list[TokenData], self.config.parse(new_tokens))
        if self.config.write(new_tokens, dry=dry) and not dry:
            self.reset()

    def reset(self):
        self.config.reset()
        try:
            data = deserialize(list[TokenData], self.config.dict)
        except ValidationError as exc:
            logger.error("Could not read API tokens", exc_info=exc)
            data = []
        self.data = {token.token: token for token in data}

    def setup(self, server):
        assert self.file
        try:
            self.config = make_config(self.file, defaults=serialize(self.defaults or []))
        except json.JSONDecodeError as exc:
            sys.exit(
                f"ERROR decoding JSON: {exc}\n"
                f"Please verify if file '{self.file}' contains valid JSON."
            )
        self.reset()

        server.app.add_middleware(
            APITokensMiddleware,
            mapper=self.get,
        )



from dataclasses import dataclass

from .page import Page
from .reg import Reference

constructors = {}


@dataclass
class BrowserEvent:
    type: str = None
    inputType: str = None
    button: int = None
    buttons: int = None
    shiftKey: bool = None
    altKey: bool = None
    ctrlKey: bool = None
    metaKey: bool = None
    key: str = None
    target: Page = None
    form: Page = None
    value: object = None
    refs: list = None
    tag: str = None

    def __getitem__(self, item):
        return getattr(self, item)

    @property
    def ref(self):
        if self.refs:
            return self.refs[0]
        else:
            return None

    @property
    def obj(self):
        return self.ref.datum


class FormData(dict):
    def __init__(self, data, target, submit, refs, tag):
        super().__init__(data)
        self.target = target
        self.submit = submit
        self.refs = refs
        self.tag = tag

    @property
    def ref(self):
        if self.refs:
            return self.refs[0]
        else:
            return None

    @property
    def obj(self):
        return self.ref.datum


class NamespaceDict(dict):
    def __getattr__(self, item):
        return self[item]


def register_constructor(key):
    def deco(fn):
        constructors[key] = fn
        return fn

    return deco


@register_constructor("HTMLElement")
def _(page, selector):
    return page[selector]


@register_constructor("Event")
def _(page, data):
    return BrowserEvent(**data)


@register_constructor("FormData")
def _(page, data, target=None, submit=None, refs=None, tag=None):
    return FormData(data, target, submit, refs, tag)


@register_constructor("Promise")
def _(page, id):
    async def resolve(value):
        await page.bearlib.resolveLocalPromise(id, value)

    return resolve


@register_constructor("Reference")
def _(page, id):
    return page.representer.object_registry.resolve(id)


@register_constructor("FullReference")
def _(page, id):
    return Reference(
        page.representer.object_registry.resolve(id),
        id=id,
    )


def construct(page, dct):
    args = dict(dct)
    name = args.pop("%")
    return constructors[name](page, **args)



from hrepr import H, J

from starbear import bear

from .utils import asset_getter

asset = asset_getter(__file__)


@bear(
    template=asset("template.html"),
    template_params={"title": "Test", "name": "Balthazar"},
    strongrefs=True,
)
async def __app__(page):
    page.add_resources(asset("stylo.css"))
    page["#box"].print(H.p["blue"]("Adding a line to the box.", id="added"))

    page.print(H.h3("Clicking the button should show 10, 20, 30, ..."))
    counter = J(module=asset("counter.js"))(increment=10, cls="c1")
    page.print(counter)
    page[counter].js.activate().__do__()

    page.print(H.h3("Clicking the button should show 0, 2, 4, ... in the blue box above it"))
    page.print(fc_target := H.div["blue"]("xxxxx", id="fancy"))
    fancy_counter = J(module=asset("fancy-counter.js"))(
        increment=2, target=page[fc_target], cls="c2"
    )
    page.print(fancy_counter)


def test_page_template(app):
    assert app.locator("title").inner_text() == "Test!"
    assert "Balthazar" in app.locator("#box").inner_text()


def test_style(app):
    added = app.locator("#added")
    assert added.evaluate("x => getComputedStyle(x).color") == "rgb(0, 0, 255)"
    assert added.evaluate("x => getComputedStyle(x).border") == "3px solid rgb(0, 0, 255)"


def test_counter(app):
    counter = app.locator(".c1")
    value0 = int(counter.inner_text())
    counter.click()
    value1 = int(counter.inner_text())
    counter.click()
    value2 = int(counter.inner_text())
    assert value2 == value1 + 10 == value0 + 20


def test_counter_initial_activation(app):
    assert app.locator(".c1").inner_text() == "10"


def test_fancy(app):
    counter = app.locator(".c2")
    target = app.locator("#fancy")
    counter.click()
    value0 = int(target.inner_text())
    counter.click()
    value1 = int(target.inner_text())
    counter.click()
    value2 = int(target.inner_text())
    assert value2 == value1 + 2 == value0 + 4



from hrepr import H
from ovld import ovld

from ..core.app import LoneBear, templates_dir


class Index(LoneBear):
    hidden = True

    def __init__(self, template=templates_dir / "index-template.html", **kwargs):
        super().__init__(self.run, template=template, **kwargs)

    async def run(self, request):
        scope = request.scope
        app = scope["app"]
        root_path = scope["root_path"]
        content = render("/", app.map, restrict=root_path)
        if content is None:
            # TODO: Not sure when that happens and what this is supposed to do
            content = render("/", app.map, restrict="/".join(root_path.split("/")[:-1]))
        return self.template(body=content or "")


def render(base_path, obj, *, restrict):
    if not base_path.startswith(restrict) and not restrict.startswith(base_path):
        return None
    return _render(base_path, obj, restrict=restrict)


@ovld
def _render(base_path: str, d: dict, *, restrict):
    def _join(p):
        return f"{base_path.rstrip('/')}{p.rstrip('/')}"

    has_results = False
    results = H.table()
    for path, value in d.items():
        real_path = _join(path) or "/"
        description = render(real_path, value, restrict=restrict)
        if description is not None:
            has_results = True
            results = results(
                H.tr(
                    H.td["url"](H.a(path, href=real_path)),
                    H.td(description),
                )
            )
    return results if has_results else None


@ovld
def _render(base_path: str, obj: object, *, restrict):
    if getattr(obj, "hidden", False):
        return None
    obj = getattr(obj, "__doc__", obj)
    return H.span("No description." if obj is None else str(obj))



from hrepr import J

from ..common import here
from ..core.constructors import register_constructor


@register_constructor("MonacoEditor")
def _(page, id, content=None, delta=None):
    store = page.representer.store
    key = ("MonacoEditor", id)
    if content is not None:
        store[key] = content
    elif key not in store:
        raise Exception("Editor value is not available from Python.")
    elif delta is None:
        return store[key]
    else:
        value = store[key]
        for offset, length, txt in delta:
            value = value[:offset] + txt + value[offset + length :]
        store[key] = value
    return store[key]


editor_module = J(namespace=here / "editor.js")

Editor = editor_module.Editor
colorized = editor_module.colorized



from typing import Any, Generator

import pytest
import uvicorn
from playwright.sync_api import Page

from starbear.server.serve import ThreadableServer


def pytest_addoption(parser):
    parser.addoption(
        "--reload-pause",
        action="store",
        default="0.5",
        type=float,
        help="How long to pause in reload tests",
    )


@pytest.fixture(scope="module")
def app_config(request):
    mb = request.module.__app__
    port = 9182
    config = uvicorn.Config(app=mb, port=port)
    server = ThreadableServer(config=config)
    with server.run_in_thread():
        yield config


@pytest.fixture
def app(page, app_config) -> Generator[Page, Any, Any]:
    page.goto("http://127.0.0.1:9182/")
    page.set_default_timeout(500)
    yield page
    if not getattr(page, "disable_error_check", False):
        assert page.locator(".bear--tabular").inner_text() == ""


@pytest.fixture(scope="module")
def app_debug_config(request):
    mb = request.module.__app__
    port = 9183
    config = uvicorn.Config(app=mb, port=port)
    server = ThreadableServer(config=config)
    with server.run_in_thread(config={"starbear": {"dev": {"debug_mode": True}}}):
        yield config


@pytest.fixture
def app_debug(page, app_debug_config) -> Generator[Page, Any, Any]:
    page.goto("http://127.0.0.1:9183/")
    page.set_default_timeout(500)
    yield page
    if not getattr(page, "disable_error_check", False):
        assert page.locator(".bear--tabular").inner_text() == ""



import time

import pytest
from hrepr import H

from starbear import bear

from .utils import asset_getter

asset = asset_getter(__file__)


@bear(strongrefs=True)
async def __app__(page):
    out = H.div("?", id="output")
    page.print(
        H.button("Set", onclick=lambda e: page[out].set(H.b("S"))),
        H.button("Print", onclick=lambda e: page[out].print(H.i("P"))),
        H.button("Clear", onclick=lambda e: page[out].clear()),
        H.button("Delete", onclick=lambda e: page[out].delete()),
        H.button(
            "Template", onclick=lambda e: page[out].template(asset("template.html"), name="deer")
        ),
        out,
    )


@pytest.fixture
def check(app):
    def fn(expected, method="text"):
        assert getattr(app.locator("#output"), f"inner_{method}")() == expected

    return fn


@pytest.fixture
def click(app):
    def fn(lbl):
        app.get_by_text(lbl).click()
        time.sleep(0.05)

    return fn


def test_set(click, check):
    click("Set")
    check("S")
    check("<b>S</b>", method="html")


def test_opseq(click, check):
    check("?")
    click("Clear")
    check("")
    click("Print")
    click("Print")
    click("Print")
    check("PPP")
    click("Set")
    check("S")
    click("Print")
    check("SP")
    check("<b>S</b><i>P</i>", method="html")


def test_delete(click, app):
    assert app.locator("#output").count() == 1
    click("Delete")
    assert app.locator("#output").count() == 0


def test_template(click, check):
    click("Template")
    check("Greetings, deer")



from dataclasses import dataclass, field

from authlib.integrations.starlette_client import OAuth as OAuthClient
from starlette.config import Config
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import RedirectResponse

from ..config import StarbearServerPlugin


class OAuthMiddleware(BaseHTTPMiddleware):
    """Gate all routes behind OAuth.

    Arguments:
        app: The application this middleware is added to.
        oauth: The OAuth object.
    """

    def __init__(self, app, oauth):
        super().__init__(app)
        self.oauth = oauth

    async def dispatch(self, request, call_next):
        if "X-API-KEY" in request.headers:
            return await call_next(request)

        if request.url.path.startswith("/_/"):
            return await call_next(request)

        user = request.session.get("user")
        if not user:
            request.session["redirect_after_login"] = str(request.url)
            return RedirectResponse(url="/_/login")
        else:
            return await call_next(request)


@dataclass
class OAuth(StarbearServerPlugin):
    name: str = None
    server_metadata_url: str = None
    client_kwargs: dict = field(default_factory=dict)
    environ: dict = field(default_factory=dict)

    def cap_require(self):
        return ["session"]

    def cap_export(self):
        return ["email"]

    async def route_login(self, request):
        redirect_uri = request.url_for("auth")
        return await self.oauth.google.authorize_redirect(request, str(redirect_uri))

    async def route_auth(self, request):
        token = await self.oauth.google.authorize_access_token(request)
        user = token.get("userinfo")
        if user:
            request.session["user"] = user
        red = request.session.get("redirect_after_login", "/")
        return RedirectResponse(url=red)

    async def route_logout(self, request):
        request.session.pop("user", None)
        return RedirectResponse(url="/")

    def setup(self, server):
        oauth_config = Config(environ=self.environ)
        oauth_module = OAuthClient(oauth_config)
        oauth_module.register(
            name=self.name,
            server_metadata_url=self.server_metadata_url,
            client_kwargs=self.client_kwargs,
        )
        server.app.add_middleware(
            OAuthMiddleware,
            oauth=oauth_module,
        )

        self.oauth = oauth_module
        server.app.add_route("/_/login", self.route_login)
        server.app.add_route("/_/logout", self.route_logout)
        server.app.add_route("/_/auth", self.route_auth, name="auth")



import asyncio
import functools
import traceback
from dataclasses import dataclass, field
from enum import Enum
from hashlib import md5
from mimetypes import guess_type

from hrepr.resource import JSExpression

from .reg import Reference

ABSENT = object()


def keyword_decorator(deco):
    """Wrap a decorator to optionally takes keyword arguments."""

    @functools.wraps(deco)
    def new_deco(fn=ABSENT, *args, **kwargs):
        if callable(fn):
            return deco(fn, *args, **kwargs)

        if fn is not ABSENT:
            args = (fn, *args)

        @functools.wraps(deco)
        def newer_deco(fn):
            return deco(fn, *args, **kwargs)

        return newer_deco

    return new_deco


class Queue(asyncio.Queue):
    def putleft(self, entry):
        self._queue.appendleft(entry)
        self._unfinished_tasks += 1
        self._finished.clear()
        self._wakeup_next(self._getters)

    def tag(self, arg):
        return ClientWrap(self).tag(arg)

    def wrap(self, **options):
        return ClientWrap(self, **options)

    def __aiter__(self):
        return self

    async def __anext__(self):
        return await self.get()

    def __call__(self, value):
        return self.put(value)


class Responses(Enum):
    NO_LISTENERS = "no_listeners"


@dataclass
class Event:
    __match_args__ = ("type", "value")

    type: str
    value: object


@dataclass
class FeedbackEvent(Event):
    response: asyncio.Future = field(default_factory=asyncio.Future)

    def resolve(self, response=None):
        self.response.set_result(response)

    def reject(self, error):
        self.response.set_exception(error)


def format_error(message, debug=None, exception=None, show_debug=False):
    if show_debug:
        tb = exception and traceback.format_exception(
            type(exception),
            exception,
            exception.__traceback__,
        )
        parts = [message, debug, tb and "\n".join(tb)]
        message = "\n\n".join(p for p in parts if p)
    return message


class FeedbackQueue(Queue):
    pass


class VirtualFile:
    def __init__(self, content, type=None, name=None):
        if type is None:
            if name is not None:
                type, _ = guess_type(url=name)

        self.type = type
        self.content = content
        self.name = md5(content.encode("utf8")).hexdigest()
        if name is not None:
            self.name += f"/{name}"


def rewrap(old_func, new_func):
    if isinstance(old_func, ClientWrap):
        return type(old_func)(new_func, **old_func.options)
    else:
        return new_func


class ClientWrap:
    FIELDS = {
        "debounce",
        "extract",
        "form",
        "refs",
        "pack",
        "partial",
        "toggles",
        "pre",
        "post",
        "tag",
    }

    def __init__(self, func, **options):
        if any((key := k) not in self.FIELDS for k in options):
            raise TypeError(f"Invalid argument to ClientWrap: '{key}'")

        if isinstance(func, ClientWrap):
            options = {**func.options, **options}
            func = func.func

        if "partial" in options:
            partial = options["partial"]
            if partial is None or isinstance(partial, (list, tuple)):
                options["partial"] = partial
            else:
                options["partial"] = [partial]

        if "debounce" in options:
            options["id"] = id(func)

        for x in ["pre", "post"]:
            result = options.get(x, None)
            if result is not None:
                if not isinstance(result, (list, tuple)):
                    result = [result]
                result = [
                    JSExpression(f"function (result) {{ {p} }}") if isinstance(p, str) else p
                    for p in result
                ]
                options[x] = result

        self.func = func
        self.options = options

    def tag(self, arg):
        return self.wrap(tag=arg if isinstance(arg, str) else Reference(arg))

    def wrap(self, **options):
        return type(self)(self, **options)

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __aiter__(self):
        return aiter(self.func)

    def __js_embed__(self, representer):
        fn = representer.js_embed(self.func)
        options = representer.js_embed(self.options)
        return f"$$BEAR.wrap({fn}, {options})"

    def __attr_embed__(self, gen):
        return f"$$BEAR.event.call(this, {gen.js_embed(self)})"



from functools import cached_property

from hrepr import H

from .functions import Multiplexer

RESET = object()


def live(obj, hrepr=None, **listeners):
    if hasattr(obj, "__live_element__"):
        elem = obj.__live_element__(H, hrepr)
    else:
        elem = H.live_element()
    return elem(runner=obj.__live__, id=True, **listeners)


class GeneratorPrinter:
    def __init__(self, generator):
        self.generator = generator


class Inplace(GeneratorPrinter):
    async def __live__(self, elem):
        async for obj in self.generator:
            elem.set(obj)


class Print(GeneratorPrinter):
    async def __live__(self, elem):
        async for obj in self.generator:
            if elem is RESET:
                elem.clear()
            else:
                elem.print(obj)


class Watchable:
    @cached_property
    def _mx(self):
        return Multiplexer()

    def notify(self, event):
        self._mx.notify(event)

    def watch_context(self):
        return self._mx.stream_context()

    async def watch(self):
        return await self._mx.stream()



import errno
import importlib
import ipaddress
import json
import os
import random
import socket
from dataclasses import dataclass, field
from functools import cached_property
from pathlib import Path
from types import ModuleType

import gifnoc
from serieux import TaggedSubclass
from serieux.model import Model, model

from ..common import UsageError
from .find import collect_locations, collect_routes, collect_routes_from_module
from .reload import (
    BaseReloader,
    FullReloader,
    InertJuriggedReloader,
    InertReloader,
    JuriggedReloader,
)


def _module_from_string(mod):
    field = None
    if ":" in mod:
        mod, field = mod.split(":", 1)
    mod = importlib.import_module(mod)
    if field is not None:
        mod = getattr(mod, field)
    return mod


@model.register
def _module_model(t: type[ModuleType]):
    return Model(
        original_type=ModuleType,
        from_string=_module_from_string,
    )


@dataclass
class StarbearSSLConfig:
    # Whether SSL is enabled
    enabled: bool = False
    # SSL key file
    keyfile: Path = None
    # SSL certificate file
    certfile: Path = None


@dataclass
class StarbearServerPlugin:
    # Whether the plugin is enabled
    enabled: bool = True

    # Whether the plugin is required (if false, only include if needed by another plugin)
    required: bool = True

    def cap_require(self):
        return []

    def cap_export(self):
        return []

    def setup(self, server):
        pass


@dataclass
class StarbearServerBaseConfig:
    # Port to serve from
    port: int = 8000

    # Hostname to serve from
    host: str = "127.0.0.1"

    # Path to watch for changes with jurigged
    watch: str = None

    # Run in development mode
    dev: bool = False

    # Automatically open browser
    open_browser: bool = False

    # Reloading methodology
    reload_mode: str = "jurigged"

    # SSL configuration
    ssl: StarbearSSLConfig = field(default_factory=StarbearSSLConfig)

    # Plugins
    plugins: dict[str, TaggedSubclass[StarbearServerPlugin]] = field(default_factory=dict)

    def __post_init__(self):
        override = os.environ.get("STARBEAR_RELOAD_OVERRIDE", None)
        if override:
            self.host, self.port = json.loads(override)
            self.open_browser = False

        if self.dev and not self.watch:
            self.watch = True
        if self.watch:
            self.dev = True

    @cached_property
    def socket(self):
        host = self.host
        if host == "127.255.255.255":
            # Generate a random loopback address (127.x.x.x)
            host = ipaddress.IPv4Address("127.0.0.1") + random.randrange(2**24 - 2)
            host = str(host)

        family = socket.AF_INET6 if ":" in host else socket.AF_INET

        sock = socket.socket(family=family)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        try:
            sock.bind((host, self.port))
        except OSError as exc:
            if self.host == "127.255.255.255" and exc.errno == errno.EADDRNOTAVAIL:
                # The full 127.x.x.x range may not be available on this system
                sock.bind(("localhost", self.port))
            else:
                raise
        return sock

    def get_reloader(self, server):
        if not self.dev:
            return InertReloader(server)
        elif self.reload_mode == "manual":
            return BaseReloader(server)
        elif self.reload_mode == "jurigged":
            return JuriggedReloader(server)
        elif self.reload_mode == "jurigged_noreload":
            return InertJuriggedReloader(server)
        else:
            return FullReloader(server)

    def get_locations(self):
        return []

    def get_routes(self):
        raise NotImplementedError("Please override get_routes()")


@dataclass
class StarbearServerConfig(StarbearServerBaseConfig):
    # Directory or script
    root: Path = None

    # Reference to the module to run
    module: ModuleType = None

    # Field to get the apps from
    module_field: str = None

    # Explicitly given routes
    routes: dict = None

    def get_locations(self):
        if self.root:
            locations = [self.root]

        elif self.module:
            locations = [Path(self.module.__file__).parent]

        elif self.routes:
            locations = list(collect_locations(self.routes))

        return [str(loc) for loc in locations]

    def get_routes(self):
        if ((self.root is not None) + (self.module is not None) + (self.routes is not None)) != 1:
            raise UsageError(
                "Either the root or module argument must be provided, or a dict of explicit routes."
            )

        if self.root:
            return collect_routes(self.root)

        elif self.module:
            return collect_routes_from_module(self.module, self.module_field)

        elif self.routes:
            return self.routes


config = gifnoc.define(
    field="starbear.server",
    model=StarbearServerConfig,
)



import os
import re
import shlex
import signal
import subprocess
from contextlib import contextmanager


@contextmanager
def serve(*args):
    # args = ["coverage", "run", "--include", "src", "-m", "starbear", "serve", "-p", "0", *map(str, args)]
    args = ["starbear", "serve", "-p", "0", *map(str, args)]
    cmd = shlex.join(args)
    print("Running command:", cmd)
    proc = subprocess.Popen(
        args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env={**os.environ, "PYTHONUNBUFFERED": "1"},
    )
    try:
        while True:
            line = proc.stderr.readline().decode("utf8")
            print(line, end="")
            if m := re.match(string=line, pattern=r".*Serving at:.*(https?://[a-zA-Z0-9_.:]+)"):
                yield m.groups()[0]
                break
            if re.match(string=line, pattern=r".*(Error|Exception).*"):
                raise Exception(f"Command failed: '{cmd}'")
    finally:
        proc.send_signal(signal.SIGINT)
        proc.wait()



from starbear import H, Queue, bear


@bear
async def __app__(page):
    def err(_):
        raise Exception("oh no!")

    q = Queue()
    page.print(
        H.button("Error 1", id="e1", onclick=q.tag("e1")),
        H.button("Error 2", id="e2", onclick=err),
    )
    async for event in q:
        if event.tag == "e1":
            1 / 0


def test_error(app):
    app.locator("#e1").click()
    txt = app.locator(".bear--tabular-area-container.bear--active").inner_text()
    assert "An error occurred" in txt
    assert "ZeroDivisionError" not in txt
    app.disable_error_check = True


def test_error_debug(app_debug):
    app_debug.locator("#e1").click()
    txt = app_debug.locator(".bear--tabular-area-container.bear--active").inner_text()
    assert "ZeroDivisionError" in txt
    app_debug.disable_error_check = True


def test_error_in_method(app):
    app.locator("#e2").click()
    txt = app.locator(".bear--tabular-area-container.bear--active").inner_text()
    assert "Application error" in txt
    assert "oh no!" not in txt
    app.disable_error_check = True


def test_error_in_method_debug(app_debug):
    app_debug.locator("#e2").click()
    txt = app_debug.locator(".bear--tabular-area-container.bear--active").inner_text()
    assert "oh no!" in txt
    app_debug.disable_error_check = True



import weakref
from collections import deque
from hashlib import md5
from itertools import count
from pathlib import Path
from types import MethodType

_c = count()


def _id(id=None):
    return next(_c) if id is None else id


class StrongRegistry:
    def __init__(self):
        self.map = {}

    def register(self, obj, id=None):
        currid = _id(id)
        self.map[currid] = obj
        return currid

    def resolve(self, id):
        return self.map[id]


class StrongRotatingRegistry:
    def __init__(self, keep, rotate):
        self.keep = keep
        self.rotate = rotate
        self.map = {}
        self.ids = deque()

    def register(self, obj, id=None):
        currid = _id(id)
        self.map[currid] = obj
        self.ids.append(currid)
        if len(self.ids) > self.keep >= 0:
            if self.rotate:
                rm = self.ids.popleft()
                del self.map[rm]
            else:
                raise Exception("Exceeded limit for keeping strong references to objects.")
        return currid

    def resolve(self, id):
        return self.map[id]


class WeakRegistry:
    def __init__(self):
        self.map = {}

    def register(self, obj, id=None):
        currid = _id(id)
        if isinstance(obj, MethodType):
            ref = weakref.WeakMethod(obj)
        else:
            ref = weakref.ref(obj)
        self.map[currid] = ref
        return currid

    def resolve(self, id):
        value = self.map[id]()
        if value is None:
            raise KeyError(id)
        else:
            return value


class ObjectRegistry:
    def __init__(self, strongrefs=100, rotate_strongrefs=True):
        self.wr = WeakRegistry()
        self.sr = StrongRotatingRegistry(keep=strongrefs, rotate=rotate_strongrefs)

    def register(self, obj, id=None):
        try:
            return self.wr.register(obj, id)
        except TypeError:
            return self.sr.register(obj, id)

    def resolve(self, id):
        try:
            return self.wr.resolve(id)
        except KeyError:
            return self.sr.resolve(id)


class Reference:
    def __init__(self, datum, id=None):
        self.id = _id(id)
        self.datum = datum


class FileRegistry:
    def __init__(self):
        self.file_to_url = {}
        self.url_to_file = {}

    def find_anchor(self, filename):
        filename = filename.absolute()
        anchor = filename
        while anchor != Path("/"):
            anchor = anchor.parent
            if anchor in self.file_to_url:
                return (anchor, self.file_to_url[anchor])
        else:
            anchor = filename.parent
            while not (anchor / "starbear-anchor").exists():
                anchor = anchor.parent
                if anchor == Path("/"):
                    anchor = filename.parent
                    break
            url = md5(str(anchor).encode("utf8")).hexdigest()
            self.file_to_url[anchor] = url
            self.url_to_file[url] = anchor
            return (anchor, url)

    def register(self, filename):
        filename = filename.absolute()
        anchor, base_url = self.find_anchor(filename)
        url = str(base_url / filename.relative_to(anchor))
        return url

    def get_file_from_url(self, url):
        url = orig = Path(url)
        while url != url.parent and str(url) not in self.url_to_file:
            url = url.parent
        anchor = self.url_to_file.get(str(url), None)
        return anchor and str(anchor / orig.relative_to(url))


class VFileRegistry:
    def __init__(self):
        self.vfiles = {}

    def register(self, vfile):
        self.vfiles[vfile.name] = vfile
        return vfile.name

    def get(self, pth):
        return self.vfiles[pth]


class FutureRegistry:
    def __init__(self):
        self.current_id = count()
        self.futures = {}

    def register(self, future):
        fid = next(self.current_id)
        self.futures[fid] = future
        return fid

    def resolve(self, fid, value):
        if fid not in self.futures:
            return
        self.futures[fid].set_result(value)
        del self.futures[fid]

    def reject(self, fid, error):
        if fid not in self.futures:
            return
        self.futures[fid].set_exception(Exception(error))
        del self.futures[fid]


class QueueRegistry:
    def __init__(self):
        self.current_id = count()
        self.queues = {}

    def register(self, queue):
        qid = next(self.current_id)
        self.queues[qid] = queue
        return qid

    def put(self, qid, value):
        self.queues[qid].put_nowait(value)



import logging
import sys
import traceback
from pathlib import Path


def _here(depth):
    fr = sys._getframe(depth)
    filename = fr.f_code.co_filename
    return Path(filename).parent


class UsageError(Exception):
    pass


class StarbearHandler(logging.StreamHandler):
    def __init__(self, channel, use_color=None):
        super().__init__(channel)
        self.use_color = sys.stdout.isatty() if use_color is None else use_color

    def colorize(self, color, text):
        if self.use_color:
            return f"\033[{color}m{text}\033[0m"
        else:
            return text

    def format(self, record):
        def _brack(s):
            return f"[{self.colorize(36, s)}]" if s else ""

        process = getattr(record, "proc", None)
        user = getattr(record, "user", None)
        tb = getattr(record, "traceback", None)
        colors = {
            "INFO": "32",
            "WARNING": "33",
            "ERROR": "31",
        }
        color = colors.get(record.levelname, "95")
        prefix = f"{self.colorize(color, record.levelname)}:   {_brack(record.name)}{_brack(process)}{_brack(user)}"
        msg = record.msg
        if tb:
            msg += "\n" + traceback.format_exc()
        if "\n" in msg:
            lines = f"\n\033[{color}m>\033[0m ".join(msg.split("\n"))
            return f"{prefix} {lines}"
        else:
            return f"{prefix} {msg}"


logger = logging.getLogger("starbear")
logger.setLevel(level=logging.INFO)
logger.addHandler(StarbearHandler(sys.stderr))


def __getattr__(attr):
    if attr == "here":
        return _here(2)

    raise AttributeError(attr)



from pathlib import Path


def asset_getter(file):
    here = Path(file)
    locations = [here.parent / "common_assets", here.parent / (here.stem + "_assets")]

    def get(name):
        for location in locations:
            candidate = location / name
            if candidate.exists():
                return candidate
        else:
            raise FileNotFoundError(name)

    return get



import time

from hrepr import H

from starbear import Queue, bear


@bear(template_params={"title": "formyform"})
async def __app__(page):
    q = Queue()
    fq = q.wrap(form=True)
    page.print(
        H.form(
            H.div(H.input(id="input1", name="input1", oninput=fq)),
            H.div(H.input(id="input2", name="input2", oninput=fq)),
            H.div(H.button("Clicky!")),
            onsubmit=fq,
        ),
        out := H.div(id="output"),
    )
    async for event in q:
        end = "!" if event.submit else "?"
        page[out].set(f"I say {event['input1']} and {event['input2']}{end}")


def test_title(app):
    assert app.locator("title").inner_text() == "formyform"


def test_manipulate(app):
    app.locator("#input1").fill("wow")
    time.sleep(0.05)
    assert app.locator("#output").inner_text() == "I say wow and ?"
    app.locator("#input2").fill("cool")
    time.sleep(0.05)
    assert app.locator("#output").inner_text() == "I say wow and cool?"
    app.locator("button").click()
    time.sleep(0.05)
    assert app.locator("#output").inner_text() == "I say wow and cool!"



from hrepr import H

from starbear import bear


@bear
async def __app__(page):
    page.print(H.div("hello world!", id="hello"))


def test_title(app):
    assert app.locator("title").inner_text() == "Starbear"


def test_charset(app):
    assert app.locator('meta[http-equiv="Content-type"]').get_attribute("charset") == "UTF-8"


def test_hello(app):
    assert app.locator("#hello").inner_text() == "hello world!"
    assert app.locator("html").inner_text() == "hello world!"  # Check there's no junk



import sys
from dataclasses import dataclass

import gifnoc
from serieux import TaggedUnion

from .common import UsageError
from .server.config import config as server_config


@dataclass
class Serve:
    """Start a Starbear server."""

    # Directory or script
    # [positional]
    path: str = None

    # Reference to the module to run
    # [alias: -m]
    module: str = None

    # Port to serve from
    # [alias: -p]
    port: int = None

    # Hostname to serve from
    host: str = None

    # Path to watch for changes with jurigged
    watch: str = None

    # Run in development mode
    # [alias: -d]
    dev: bool = None

    # Automatically open browser
    browser: bool = None

    # Reloading methodology
    reload_mode: str = None

    # SSL key file
    ssl_keyfile: str = None

    # SSL certificate file
    ssl_certfile: str = None

    def __call__(self):
        cfg = {
            "starbear.server.root": self.path,
            "starbear.server.module": self.module,
            "starbear.server.port": self.port,
            "starbear.server.host": self.host,
            "starbear.server.ssl.keyfile": self.ssl_keyfile,
            "starbear.server.ssl.certfile": self.ssl_certfile,
            "starbear.server.dev": self.dev,
            "starbear.server.reload_mode": self.reload_mode,
            "starbear.server.watch": self.watch,
            "starbear.server.open_browser": self.browser,
        }
        cfg = {k: v for k, v in cfg.items() if v is not None}

        gifnoc.add_overlay(cfg)

        from .server.serve import StarbearServer

        try:
            server = StarbearServer(server_config)
            server.run()
        except UsageError as exc:  # pragma: no cover
            exit(f"ERROR: {exc}")
        except FileNotFoundError as exc:  # pragma: no cover
            exit(f"ERROR: File not found: {exc}")


def main(argv=None):
    sys.path.append(".")

    command = gifnoc.cli(
        TaggedUnion[Serve],
        mapping={"starbear.dev.debug_mode": "--debug"},
        argv=argv or sys.argv[1:],
    )
    command()


if __name__ == "__main__":  # pragma: no cover
    main()



from dataclasses import dataclass, field
from typing import Any

import gifnoc


@dataclass
class StarbearDevConfig:
    # Turn on debug mode
    debug_mode: bool = False

    # Dev code to inject
    inject: list[Any] = field(default_factory=list)


@dataclass
class StarbearConfig:
    dev: StarbearDevConfig = field(default_factory=StarbearDevConfig)


config = gifnoc.define(
    field="starbear",
    model=StarbearConfig,
    defaults={},
)



import requests

from starbear import simplebear

from .utils import asset_getter

asset = asset_getter(__file__)


async def exponential(x):
    return (x * 2) or 1


@simplebear
async def __app__(request):
    x = request.query_params["x"]
    return {"square": int(x) ** 2}


def test_app(app_config):
    resp = requests.get("http://127.0.0.1:9182/?x=12")
    assert resp.json() == {"square": 144}

    resp = requests.get("http://127.0.0.1:9182/?x=7")
    assert resp.json() == {"square": 49}



import json
import sys
from collections import defaultdict
from dataclasses import dataclass
from fnmatch import fnmatch
from pathlib import Path
from typing import Optional

from hrepr import H
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import HTMLResponse

from ...common import UsageError
from ..config import StarbearServerPlugin


class PermissionsMiddleware(BaseHTTPMiddleware):
    def __init__(self, app, is_authorized):
        super().__init__(app)
        self.router = app
        self.is_authorized = is_authorized

    async def dispatch(self, request, call_next):
        if (path := request.url.path).startswith("/_/"):
            return await call_next(request)

        user = request.session.get("user")
        if not self.is_authorized(user, path):
            content = H.body(
                H.h2("Forbidden"),
                H.p("User ", H.b(user["email"]), " cannot access this page."),
                H.a("Logout", href="/_/logout"),
            )
            return HTMLResponse(str(content), status_code=403)
        else:
            return await call_next(request)


@dataclass
class Permissions(StarbearServerPlugin):
    # Configuration file in which the permissions are located
    file: Optional[Path] = None

    # Default permissions to bootstrap the configuration file
    defaults: Optional[dict] = None

    def cap_require(self):
        return ["email"]

    def cap_export(self):
        return []

    def setup(self, server):
        if self.file:
            try:
                permissions = PermissionFile(self.file, defaults=self.defaults)
            except json.JSONDecodeError as exc:
                sys.exit(
                    f"ERROR decoding JSON: {exc}\n"
                    f"Please verify if file '{self.file}' contains valid JSON."
                )
        elif self.defaults:
            permissions = PermissionDict(self.defaults)
        else:
            raise UsageError("The permissions plugin should specify a file and/or defaults.")

        self.permissions = permissions
        server.app.add_middleware(PermissionsMiddleware, is_authorized=permissions)


class ConfigFile:
    def __init__(self, file, defaults=None):
        self.dict = None
        self.file = file
        if not self.file.exists():
            if defaults is None:
                raise FileNotFoundError(self.file)
            else:
                self.write(defaults)
        self.reset()

    def reset(self):
        self.dict = self.parse(self.read())

    def read(self):
        return self.file.read_text()

    def write(self, new_content, dry=False):
        if not isinstance(new_content, str):
            new_content = self.unparse(new_content)
        if dry:
            # Check that the new content is valid
            self.parse(new_content)
        else:
            if self.file.exists():
                previous = self.read()
            else:
                previous = self.unparse({})
            self.file.write_text(new_content)
            try:
                self.reset()
            except Exception:
                self.file.write_text(previous)
                self.reset()
                raise
        return True


class JSONFile(ConfigFile):
    def parse(self, content):
        return json.loads(content)

    def unparse(self, data):
        return json.dumps(data)


class YAMLFile(ConfigFile):
    def parse(self, content):
        import yaml

        return yaml.safe_load(content)

    def unparse(self, data):
        import yaml

        return yaml.safe_dump(data)


extensions_map = {
    ".json": JSONFile,
    ".yaml": YAMLFile,
    ".yml": YAMLFile,
}


def make_config(config_file, defaults=None):
    config_file = Path(config_file)
    suffix = config_file.suffix
    cls = extensions_map.get(suffix, None)
    if cls is None:
        raise UsageError(f"Unknown config file extension: {suffix}")
    else:
        return cls(config_file, defaults=defaults)


def parse_config():
    return make_config().dict


class PermissionDict:
    def __init__(self, permissions):
        self.permissions = permissions
        self.reset()

    def reset(self):
        self.cache = defaultdict(dict)
        self.wild = defaultdict(list)
        for path, allowed in self.permissions.items():
            if path == "/":
                path = ("",)
            else:
                path = tuple(path.split("/"))
            for user in allowed:
                if "*" in user:
                    self.wild[path].append(user)
                else:
                    self.cache[path][user] = True

    def __call__(self, user, path):
        path, *_ = path.split("!", 1)
        path = path.removesuffix("/")
        if path == "/":
            parts = ("",)
        else:
            parts = tuple(path.split("/"))
        email = user["email"]
        partials = [(*parts[: i + 1], "**") for i in range(len(parts))]
        to_check = [parts, *partials]
        for current in to_check:
            cache = self.cache[current]
            if email in cache:
                if cache[email]:
                    return True
            else:
                for wild in self.wild[current]:
                    if fnmatch(email, wild):
                        cache[email] = True
                        return True
                else:
                    cache[email] = False
        return False


class PermissionFile(PermissionDict):
    def __init__(self, permissions_file, defaults=None):
        self.file = make_config(permissions_file, defaults=defaults)
        super().__init__(self.file.dict)

    def reset(self):
        self.permissions = self.file.dict
        super().reset()

    def read(self):
        return self.file.read()

    def write(self, new_permissions, dry=False):
        if self.file.write(new_permissions, dry=dry) and not dry:
            self.reset()



from hrepr import H, Tag

from starbear import template


def test_template_int():
    tpl = "<div>{{x}}</div>"
    result = template(tpl, x=3)
    assert isinstance(result, Tag)
    assert str(result) == "<div>3</div>"


def test_template_tag():
    tpl = "<div>{{x}}</div>"
    result = template(tpl, x=H.b("hello"))
    assert str(result) == "<div><b>hello</b></div>"



from .config import StarbearServerBaseConfig, StarbearServerConfig, StarbearServerPlugin
from .serve import StarbearServer

__all__ = [
    "StarbearServer",
    "StarbearServerBaseConfig",
    "StarbearServerConfig",
    "StarbearServerPlugin",
]



from starlette.responses import HTMLResponse
from starlette.routing import Route

from starbear import H, bear
from starbear.server.misc import simple_route


@bear
async def jupiter(page):
    """Planet Jupiter"""
    page.print(H.h1("JUPITER!"))


@bear
async def titan(page):
    """Moon titan"""
    page.print(H.h1("TITAN!"))


@simple_route
async def europa(request):
    return HTMLResponse("<h1>EUROPA!</h1>")


async def ganymede(request):
    return HTMLResponse("<h1>GANYMEDE!</h1>")


__app__ = {
    "/index": jupiter,
    "/titan": titan,
    "/europa": europa,
    "/ganymede": Route("/ganymede", endpoint=ganymede),
}



import pytest
from hrepr import H

from starbear import bear
from starbear.components.editor import Editor, colorized

editor_text = "def f(x):\n    return x * x"


@bear
async def __app__(page):
    def f(change):
        if change["event"] == "change":
            page.print(H.div(change["content"]))
        else:
            page.print(change)

    page.print(
        H.div(
            H.h1("Colorized text"),
            H.div["colorized"](colorized(text=editor_text, language="python")),
            H.h1("Interactive editor"),
            H.div["editor"](
                Editor(
                    value=editor_text,
                    autofocus=True,
                    language="python",
                    onChange=f,
                    bindings={"WinCtrl+KeyA": f},
                )
            ),
        )
    )
    await page.wait()


@pytest.fixture
def ed(app):
    app.set_default_timeout(2500)  # monaco takes time to load
    yield app.locator(".editor .starbear-editor-area")


def test_colorized(app):
    app.set_default_timeout(2500)  # monaco takes time to load
    assert app.locator(":nth-match(.colorized .mtk6, 1)").inner_text() == "def"
    assert app.locator(":nth-match(.colorized .mtk6, 2)").inner_text() == "return"


def test_editor_value(ed):
    assert ed.evaluate("async x => (await x.__object).editor.getValue()") == editor_text


def test_editor_autofocus(ed):
    assert ed.evaluate("x => x.contains(document.activeElement)")



from starlette.routing import Route


def simple_route(fn=None, *, route_class=Route, **kwargs):
    def wrap(fn):
        fn.route_class = route_class
        fn.route_parameters = kwargs
        return fn

    return wrap if fn is None else wrap(fn)



import pytest

from starbear import here

from .utils import serve


@pytest.fixture(scope="module")
def app_hello():
    with serve("-m", "tests.server.app_hello") as addr:
        yield addr


def test_index(app_hello, page):
    page.goto(f"{app_hello}")
    text = page.locator("body").inner_text()
    assert "hello" in text
    assert "world" in text
    assert "Venus" in text
    assert "Jupiter" in text
    assert "Mars" in text


def test_subindex(app_hello, page):
    page.goto(f"{app_hello}/world")
    text = page.locator("body").inner_text()
    assert "hello" not in text
    assert "world" in text
    assert "Venus" in text
    assert "Jupiter" in text
    assert "Mars" in text


def test_hello(app_hello, page):
    page.goto(f"{app_hello}/hello")
    assert page.locator("h1").inner_text() == "HELLO"


def test_subdirectory(app_hello, page):
    page.goto(f"{app_hello}/world/venus")
    assert page.locator("h1").inner_text() == "VENUS!"


def test_root_argument(page):
    page.set_default_timeout(500)
    with serve(here / "app_hello" / "world" / "venus.py") as addr:
        page.goto(addr)
        assert page.locator("h1").inner_text() == "VENUS!"


def test_subroutes_index(app_hello, page):
    page.goto(f"{app_hello}/world/jupiter")
    assert page.locator("h1").inner_text() == "JUPITER!"


def test_subroutes_bear(app_hello, page):
    page.goto(f"{app_hello}/world/jupiter/titan")
    assert page.locator("h1").inner_text() == "TITAN!"


def test_subroutes_simple_route(app_hello, page):
    page.goto(f"{app_hello}/world/jupiter/europa")
    assert page.locator("h1").inner_text() == "EUROPA!"


def test_subroutes_starlette_route(app_hello, page):
    page.goto(f"{app_hello}/world/jupiter/ganymede")
    assert page.locator("h1").inner_text() == "GANYMEDE!"



from dataclasses import dataclass

from ..config import StarbearServerPlugin


@dataclass
class Sentry(StarbearServerPlugin):
    dsn: str = None
    traces_sample_rate: float = None
    environment: str = None
    log_level: str = None
    event_log_level: str = None

    def cap_require(self):
        return []

    def cap_export(self):
        return []

    def setup(self, server):
        import logging

        import sentry_sdk

        # Configure sentry to collect log events with minimal level INFO
        # (2023/10/25) https://docs.sentry.io/platforms/python/integrations/logging/
        from sentry_sdk.integrations.logging import LoggingIntegration

        def _get_level(level_name: str) -> int:
            level = logging.getLevelName(level_name)
            return level if isinstance(level, int) else logging.INFO

        sentry_sdk.init(
            dsn=self.dsn,
            traces_sample_rate=self.traces_sample_rate,
            environment=self.environment,
            integrations=[
                LoggingIntegration(
                    level=_get_level(self.log_level or "ERROR"),
                    event_level=_get_level(self.event_log_level or "ERROR"),
                )
            ],
        )



from asyncio import Future, Queue
from dataclasses import dataclass, field, fields as dataclass_fields, is_dataclass
from pathlib import Path
from types import AsyncGeneratorType, FunctionType, MethodType
from typing import Union

from hrepr import BlockGenerator, HTMLGenerator, Interface, StdHrepr, config_defaults
from hrepr.hgen import HasNodeName
from ovld import call_next, extend_super

from ..stream.live import Inplace, live
from .reg import (
    FileRegistry,
    FutureRegistry,
    ObjectRegistry,
    QueueRegistry,
    Reference,
    StrongRegistry,
    VFileRegistry,
    WeakRegistry,
)
from .utils import FeedbackQueue, VirtualFile


class StarbearHrepr(StdHrepr):
    @extend_super
    def hrepr(self, obj: object):
        if hasattr(obj, "__live__"):
            return live(obj, hrepr=self)
        elif hasattr(obj, "__hrepr__"):
            return obj.__hrepr__(self.H, self)
        elif is_dataclass(type(obj)):
            return self.make.instance(
                title=type(obj).__name__,
                fields=[[field.name, getattr(obj, field.name)] for field in dataclass_fields(obj)],
                delimiter="=",
                type=type(obj),
            )
        else:
            return NotImplemented

    def hrepr(self, obj: AsyncGeneratorType):  # noqa: F811
        return live(Inplace(obj), hrepr=self)


hrepr = Interface(StarbearHrepr, **config_defaults)


class RepresenterState:
    def __init__(self, route, strongrefs=False):
        self.route = route
        self.store = {}
        if strongrefs is True:
            self.object_registry = StrongRegistry()
        elif not strongrefs:
            self.object_registry = WeakRegistry()
        elif strongrefs < 0:
            self.object_registry = ObjectRegistry(strongrefs=-strongrefs, rotate_strongrefs=True)
        else:
            self.object_registry = ObjectRegistry(strongrefs=strongrefs, rotate_strongrefs=False)
        self.file_registry = FileRegistry()
        self.vfile_registry = VFileRegistry()
        self.future_registry = FutureRegistry()
        self.queue_registry = QueueRegistry()


class StarbearHTMLGenerator(HTMLGenerator):
    def __init__(self, representer_state):
        self.state = representer_state
        super().__init__(block_generator_class=StarbearBlockGenerator, hrepr=hrepr)


@dataclass
class StarbearBlockGenerator(BlockGenerator):
    live_generators: dict = field(default_factory=dict)

    @property
    def route(self):
        return self.global_generator.state.route

    def register_object(self, x, **kwargs):
        return self.global_generator.state.object_registry.register(x, **kwargs)

    def register_file(self, x, **kwargs):
        return self.global_generator.state.file_registry.register(x, **kwargs)

    def register_vfile(self, x, **kwargs):
        return self.global_generator.state.vfile_registry.register(x, **kwargs)

    def register_future(self, x, **kwargs):
        return self.global_generator.state.future_registry.register(x, **kwargs)

    def register_queue(self, x, **kwargs):
        return self.global_generator.state.queue_registry.register(x, **kwargs)

    @extend_super
    def node_embed(self, elem: HasNodeName["live-element"]):  # noqa: F811, F821
        attrs = dict(elem.attributes)
        runner = attrs["runner"]
        attrs["runner"] = False
        listeners = {}
        for k, v in list(attrs.items()):
            if k == "on-produce":
                listeners[True] = v
                attrs[k] = False
            elif k.startswith("on-produce-"):
                listeners[k[11:]] = v
                attrs[k] = False
        self.live_generators[elem.id] = (runner, listeners)
        return call_next(elem.fill(attributes=attrs))

    @extend_super
    def js_embed(self, fn: Union[MethodType, FunctionType]):  # noqa: F811
        method_id = self.register_object(fn)
        return f"$$BEAR.func({method_id})"

    def js_embed(self, ref: Reference):  # noqa: F811
        obj_id = self.register_object(ref.datum)
        return f"$$BEAR.ref({obj_id})"

    def js_embed(self, pth: Path):  # noqa: F811
        new_pth = self.register_file(pth)
        return f"'{self.route}/file/{new_pth}'"

    def js_embed(self, future: Future):  # noqa: F811
        fid = self.register_future(future)
        return f"$$BEAR.promise({fid})"

    def js_embed(self, queue: Queue):  # noqa: F811
        qid = self.register_queue(queue)
        return f"$$BEAR.queue({qid})"

    def js_embed(self, queue: FeedbackQueue):  # noqa: F811
        qid = self.register_queue(queue)
        return f"$$BEAR.queue({qid}, true)"

    @extend_super
    def attr_embed(self, fn: Union[MethodType, FunctionType]):  # noqa: F811
        method_id = self.register_object(fn)
        return f"$$BEAR.event.call(this, $$BEAR.func({method_id}))"

    def attr_embed(self, ref: Reference):  # noqa: F811
        obj_id = self.register_object(ref.datum, id=ref.id)
        return f"obj#{obj_id}"

    def attr_embed(self, queue: Queue):  # noqa: F811
        qid = self.register_queue(queue)
        return f"$$BEAR.event.call(this, $$BEAR.queue({qid}))"

    def attr_embed(self, pth: Path):  # noqa: F811
        new_pth = self.register_file(pth)
        return f"{self.route}/file/{new_pth}"

    def attr_embed(self, vf: VirtualFile):  # noqa: F811
        pth = self.register_vfile(vf)
        return f"{self.route}/vfile/{pth}"



from dataclasses import dataclass

from hrepr import H

from starbear import Queue, Reference, bear

from .utils import asset_getter

asset = asset_getter(__file__)


@dataclass
class Person:
    name: str
    job: str
    toggle: bool = True


people = [
    Person(name="Olivier", job="Programmer"),
    Person(name="Balthazar", job="Mage"),
    Person(name="Elizabeth", job="Queen"),
]


@bear
async def __app__(page):
    q = Queue()
    for person in people:
        page.print(
            H.div[person.name](
                H.div["toggle-area"](person.name, cls=[person.name]),
                H.button("Toggle", onclick=q.wrap(refs=True)),
                __ref=Reference(person),
            )
        )

    async for event in q:
        person = event.obj
        person.toggle = not person.toggle
        page[event.ref, ".toggle-area"].set(person.name if person.toggle else person.job)


def test_toggles(app):
    def check(name, expected):
        assert app.locator(f".{name} .toggle-area").inner_text() == expected

    check("Olivier", "Olivier")
    check("Balthazar", "Balthazar")
    check("Elizabeth", "Elizabeth")

    app.locator(".Balthazar button").click()
    check("Olivier", "Olivier")
    check("Balthazar", "Mage")
    check("Elizabeth", "Elizabeth")

    app.locator(".Balthazar button").click()
    check("Balthazar", "Balthazar")

    app.locator(".Olivier button").click()
    app.locator(".Balthazar button").click()
    app.locator(".Elizabeth button").click()
    check("Olivier", "Programmer")
    check("Balthazar", "Mage")
    check("Elizabeth", "Queen")



from dataclasses import dataclass
from uuid import uuid4

from starlette.middleware.sessions import SessionMiddleware

from ..config import StarbearServerPlugin


@dataclass
class Session(StarbearServerPlugin):
    def cap_require(self):
        return []

    def cap_export(self):
        return ["session"]

    def setup(self, server):
        server.app.add_middleware(SessionMiddleware, secret_key=uuid4().hex)



"""Find spacebears/routes in a file or directory"""

import importlib
import pkgutil
import runpy
import sys
from functools import reduce
from pathlib import Path

import starlette
from ovld import ovld
from starlette.routing import Mount, Route

import starbear

from ..core.app import AbstractBear
from .index import Index


def collect_routes_from_module(mod, module_field=None):
    def process_module(path, submod, ispkg):
        subroutes = getattr(submod, module_field or "__app__", None)
        if subroutes is not None:
            routes[path] = subroutes
        elif ispkg:
            routes[path] = collect_routes_from_module(submod)

    locations = mod.__spec__.submodule_search_locations
    routes = {}
    if locations is None:
        process_module("/", mod, False)
    else:
        for info in pkgutil.iter_modules(locations, prefix=f"{mod.__name__}."):
            submod = importlib.import_module(info.name)
            path = f"/{submod.__name__.split('.')[-1]}/"
            process_module(path, submod, info.ispkg)

    if "/index/" not in routes.keys():
        routes["/index/"] = Index()

    return routes


def _flatten(routes):
    return reduce(list.__iadd__, routes, [])


def _mount(path, routes):
    if path == "/":
        return routes
    else:
        return [Mount(path, routes=routes)]


def collect_routes(path):
    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"Cannot find route from non-existent path: {path}")

    if path.is_dir():
        sys.path.append(str(path.parent))
        mod = importlib.import_module(path.stem)
        return {"/": collect_routes_from_module(mod)}

    else:
        glb = runpy.run_path(str(path))
        return {"/": glb["__app__"]}


@ovld
def compile_routes(path, routes: dict):
    routes = {pth.rstrip("/"): r for pth, r in routes.items()}
    if "/" not in routes:
        if "/index" in routes:
            routes["/"] = routes["/index"]
        else:
            routes["/"] = Index()
    return _mount(
        path,
        _flatten([compile_routes(path2, route) for path2, route in routes.items()]),
    )


@ovld
def compile_routes(path, mb: AbstractBear):  # noqa: F811
    return _mount(path, mb.routes())


@ovld
def compile_routes(path, obj: object):  # noqa: F811
    # TODO: better error when someone forgets @bear
    if callable(obj):
        cls = getattr(obj, "route_class", Route)
        route_parameters = getattr(obj, "route_parameters", {})
        return [cls(path, obj, **route_parameters)]
    else:
        raise TypeError(f"Cannot compile route for {path}: {obj}")


@ovld
def compile_routes(path, obj: Route):  # noqa: F811
    assert obj.path == path
    return [obj]


exclusions = {
    Path(starbear.__file__).parent,
    Path(starlette.__file__).parent,
}


@ovld
def collect_locations(routes: dict):  # noqa: F811
    rval = set()
    for subroutes in routes.values():
        rval.update(collect_locations(subroutes))
    return rval


@ovld
def collect_locations(b: AbstractBear):  # noqa: F811
    return collect_locations(getattr(b, "fn", None))


@ovld
def collect_locations(obj: object):  # noqa: F811
    if hasattr(obj, "__globals__"):
        loc = obj.__globals__.get("__file__", None)
        if not loc or any(Path(loc).is_relative_to(x) for x in exclusions):
            return set()
        return {Path(loc).parent}
    elif hasattr(obj, "__call__"):
        return collect_locations(obj.__call__)
    else:
        return set()



import asyncio
import time
from dataclasses import dataclass

from hrepr import H

from starbear.core.app import bear
from starbear.core.utils import FeedbackEvent, Queue, Responses
from starbear.stream.live import live

base_delay = 0.1


@dataclass
class Teddy:
    value: str

    async def __live__(self, element):
        element.print(self.value)


@dataclass
class OOBPrinter:
    value: str
    delay: float
    target: str

    async def __live__(self, element):
        await asyncio.sleep(self.delay)
        element.page_select(self.target).print(self.value)


@dataclass
class Producer:
    alpha: int

    async def __live__(self, element):
        element.print(
            H.div["good"]("good: "),
            H.div["bad"]("bad: "),
            H.div["unknown"]("unknown: "),
        )
        for i in range(10):
            etype = "odd" if i % 2 == 1 else "even"
            goodness = await (yield FeedbackEvent(etype, self.alpha**i))
            if goodness is True:
                element[".good"].print(H.b(i))
            elif goodness is False:
                element[".bad"].print(H.b(i))
            elif goodness is Responses.NO_LISTENERS:
                element[".unknown"].print(H.b(i))


@bear
async def __app__(page):
    q = Queue()
    page.print(container := H.div(id=True))
    page.print(H.div(id="out-of-band1"))
    page.print(H.div(id="out-of-band2"))
    page.print(H.div(Teddy("hello"), id="teddy"))
    # The first OOBPrinter is erased by the second, which means that it should be
    # cancelled. Might be fiddly with the delays.
    page[container].set(
        OOBPrinter(value="should not happen", delay=2 * base_delay, target="#out-of-band1")
    )
    page[container].set(OOBPrinter(value="yay!", delay=base_delay, target="#out-of-band2"))
    page.print(live(Producer(2), on_produce_odd=q))
    async for event in q:
        assert event.type == "odd"
        event.resolve(event.value == 32)


def test_teddy(app):
    assert app.locator("#teddy").inner_text() == "hello"


def test_delayed(app):
    # Try increasing base_delay if this fails
    time.sleep(3 * base_delay)
    assert app.locator("#out-of-band1").inner_text() == ""
    assert app.locator("#out-of-band2").inner_text() == "yay!"


def test_events(app):
    assert app.locator(".good").inner_text() == "good: 5"
    assert app.locator(".bad").inner_text() == "bad: 1379"
    assert app.locator(".unknown").inner_text() == "unknown: 02468"



from pathlib import Path

import pytest

from starbear.server.find import (
    collect_locations,
    collect_routes,
    collect_routes_from_module,
    compile_routes,
)


def test_compile_routes_error():
    with pytest.raises(TypeError):
        compile_routes("/", {"/": "???"})


def test_collect_routes_error():
    with pytest.raises(FileNotFoundError):
        collect_routes("nonexistent")


def test_collect_locations():
    from . import app_hello

    locs = collect_locations(collect_routes_from_module(app_hello))
    assert all(x.is_relative_to(Path(app_hello.__file__).parent) for x in locs)



import asyncio
import inspect
import math
import time
from contextlib import asynccontextmanager
from itertools import count as _count

from ..core.utils import Queue


class MergeStream:
    def __init__(self, *streams, stay_alive=False):
        self.queue = Queue()
        self.active = 1 if stay_alive else 0
        for stream in streams:
            self.add(stream)

    async def _add(self, fut, iterator):
        try:
            result = await fut
            self.queue.put_nowait((result, iterator))
        except StopAsyncIteration:
            self.queue.put_nowait((None, False))

    def add(self, fut):
        self.active += 1
        if inspect.isasyncgen(fut) or hasattr(fut, "__aiter__"):
            it = aiter(fut)
            coro = self._add(anext(it), it)
        else:
            coro = self._add(fut, None)
        return asyncio.create_task(coro)

    async def __aiter__(self):
        async for result, it in self.queue:
            if it is False:
                self.active -= 1
            elif it is None:
                yield result
                self.active -= 1
            else:
                asyncio.create_task(self._add(anext(it), it))
                yield result
            if self.active == 0:
                break


DONE = object()


class Multiplexer:
    def __init__(self, source=None):
        self.source = source
        self.queues = set()
        self.done = False
        self._is_hungry = asyncio.Future()
        if source is not None:
            self.main_coroutine = asyncio.create_task(self.run())

    def notify(self, event):
        for q in self.queues:
            q.put_nowait(event)

    def end(self):
        assert not self.main_coroutine
        self.done = True
        self.notify(DONE)

    def _be_hungry(self):
        if not self._is_hungry.done():
            self._is_hungry.set_result(True)

    @asynccontextmanager
    async def stream_context(self):
        q = Queue()
        self.queues.add(q)
        try:
            yield q
        finally:
            self.queues.discard(q)

    async def stream(self):
        if self.done:
            return
        self._be_hungry()
        async with self.stream_context() as q:
            async for event in q:
                if event is DONE:
                    break
                if q.empty:
                    self._be_hungry()
                yield event

    async def run(self):
        async for event in self.source:
            await self._is_hungry
            self._is_hungry = asyncio.Future()
            self.notify(event)
        self.main_coroutine = None
        self.end()

    def __hrepr__(self, H, hrepr):
        return hrepr(self.stream())


merge = MergeStream


async def repeat(value_or_func, *, count=None, interval):
    i = 0
    if count is None:
        count = math.inf
    while i < count:
        if callable(value_or_func):
            yield value_or_func()
        else:
            yield value_or_func
        await asyncio.sleep(interval)
        i += 1


async def count(interval):
    for i in _count():
        yield i
        await asyncio.sleep(interval)


async def take(stream, n):
    curr = 0
    async for x in stream:
        yield x
        curr += 1
        if curr >= n:
            break


async def filter(stream, fn):
    async for x in stream:
        if fn(x):
            yield x


async def map(stream, fn):
    async for x in stream:
        yield fn(x)


async def scan(stream, fn, init=None):
    current = init
    async for x in stream:
        current = fn(current, x)
        yield current


async def debounce(stream, delay=None, max_wait=None):
    MARK = object()

    async def mark(delay):
        await asyncio.sleep(delay)
        return MARK

    ms = MergeStream()
    max_time = None
    target_time = None
    ms.add(stream)
    current = None
    async for element in ms:
        now = time.time()
        if element is MARK:
            delta = target_time - now
            if delta > 0:
                ms.add(mark(delta))
            else:
                yield current
                max_time = None
                target_time = None
        else:
            new_element = target_time is None
            if max_time is None and max_wait is not None:
                max_time = now + max_wait
            target_time = now + delay
            if max_time:
                target_time = min(max_time, target_time)
            if new_element:
                ms.add(mark(target_time - now))
            current = element



from hrepr import H, J, returns
from hrepr.resource import Resource

from .common import UsageError, _here
from .config import config
from .core.app import bear, simplebear
from .core.constructors import BrowserEvent, FormData, NamespaceDict, register_constructor
from .core.page import Component, Page, selector_for
from .core.reg import Reference
from .core.repr import hrepr
from .core.templating import Template, template
from .core.utils import (
    ClientWrap,
    Event,
    FeedbackEvent,
    FeedbackQueue,
    Queue,
    Responses,
    VirtualFile,
    rewrap,
)
from .stream.live import GeneratorPrinter, Inplace, Print, Watchable, live
from .version import version

__all__ = [
    "H",
    "J",
    "returns",
    "Resource",
    "UsageError",
    "here",
    "config",
    "bear",
    "simplebear",
    "BrowserEvent",
    "FormData",
    "NamespaceDict",
    "register_constructor",
    "Component",
    "Page",
    "selector_for",
    "Reference",
    "hrepr",
    "live",
    "Inplace",
    "GeneratorPrinter",
    "Print",
    "Watchable",
    "Template",
    "template",
    "ClientWrap",
    "Event",
    "FeedbackEvent",
    "FeedbackQueue",
    "Queue",
    "Responses",
    "VirtualFile",
    "rewrap",
    "version",
]


def __getattr__(attr):
    if attr == "here":
        return _here(2)

    raise AttributeError(attr)



import re
from pathlib import Path
from typing import Union

from hrepr import H, Tag
from lxml import etree
from ovld import ovld


class Placeholder:
    def __init__(self, type, name):
        self.type = type
        self.name = name

    def __str__(self):
        return "{{" + self.type + ":" + self.name + "}}"

    __repr__ = __str__


class PlaceholderSequence(list):
    pass


def _parse_template(path_or_string):
    if isinstance(path_or_string, Path):
        source = path_or_string.read_text()
    elif isinstance(path_or_string, str):
        source = path_or_string
    else:
        raise TypeError("path_or_string argument should be a Path or a str")

    html = etree.fromstring(source)
    return _html_to_h(html)


def _extract_placeholders(txt, single=False):
    parts = re.split(string=txt, pattern=r"\{\{ *([a-zA-Z0-9_]* *:?[^}]+) *\}\}")
    results = []
    for i, p in enumerate(parts):
        if i % 2 == 0:
            if p:
                results.append(p)
        else:
            if ":" in p:
                typ, name = p.split(":")
            else:
                typ = "variable"
                name = p
            results.append(Placeholder(type=typ, name=name))
    if single:
        if not results:
            return ""
        elif len(results) == 1:
            return results[0]
        else:
            return PlaceholderSequence(results)
    return results


def _html_to_h(etree):
    tag = etree.tag
    if "}" in tag:
        base_node = getattr(H, tag.split("}")[1])
    else:
        base_node = getattr(H, tag)
    children = []
    if etree.text:
        children.extend(_extract_placeholders(etree.text))
    for subtree in etree:
        children.append(_html_to_h(subtree))
        if subtree.tail:
            children.append(_extract_placeholders(subtree.tail))
    return base_node.fill(
        attributes=dict(
            (x, True if y == "" else _extract_placeholders(y, single=True))
            for x, y in etree.items()
        ),
        children=children,
    )


_cached_templates = {}


class Template:
    def __init__(self, tpl, location=None):
        self.location = location
        if isinstance(tpl, Tag):
            self.template = tpl
        elif isinstance(tpl, Path):
            self.location = location or tpl.parent
            tpl = _parse_template(tpl)
        elif isinstance(tpl, str):
            tpl = _parse_template(tpl)
        self.template = tpl

    def __call__(self, **values):
        values.setdefault("_variable", lambda name: values[name])
        if self.location:

            def embed(name):
                return template(
                    self.location / name,
                    **values,
                )

            values.setdefault("_embed", embed)
        return _template(self.template, values)


def template(tpl, nocache=False, /, **values):
    if isinstance(tpl, Tag):
        return Template(tpl)(**values)
    if tpl not in _cached_templates or nocache:
        _cached_templates[tpl] = Template(tpl)
    return _cached_templates[tpl](**values)


@ovld
def _template(nodes: Union[list, tuple], values: dict):
    return type(nodes)(_template(node, values) for node in nodes)


@ovld
def _template(seq: PlaceholderSequence, values: dict):
    return "".join(map(str, (_template(x, values) for x in seq)))


@ovld
def _template(node: Placeholder, values: dict):
    processor = values[f"_{node.type}"]
    return _template(processor(node.name), values)


@ovld
def _template(node: Tag, values: dict):
    return Tag(node.name).fill(
        children=[_template(child, values) for child in node.children],
        attributes={k: _template(v, values) for k, v in node.attributes.items()},
    )


@ovld
def _template(tpl: Template, values: dict):
    return tpl(**values)


@ovld
def _template(node: object, values: dict):
    return node



import asyncio as aio
import inspect
from pathlib import Path

from hrepr import H, J, Tag
from hrepr.textgen import Breakable, Sequence

from ..common import logger
from .reg import Reference
from .repr import StarbearHTMLGenerator
from .utils import Event, FeedbackEvent, Queue, Responses, format_error


def selector_for(x):
    if isinstance(x, Tag):
        if not x.id:
            raise Exception("Cannot locate element because it has no id.")
        return f"#{x.id}"
    elif isinstance(x, J):
        return f"#{x._get_id()}"
    elif isinstance(x, str):
        return x
    elif isinstance(x, Reference):
        return f'[--ref="obj#{x.id}"]'
    elif (selector := getattr(x, "selector", None)) and isinstance(selector, str):
        return selector
    else:
        raise TypeError(
            "A valid selector must be a str, Tag, Reference or an object"
            " with a string attribute named `selector`."
        )


class Component:
    @property
    def selector(self):
        return selector_for(self.node)

    def __hrepr__(self, H, hrepr):
        return self.node

    def __h__(self):
        return self.node


class AwaitableJ(J):
    def __init__(self, page=None, **kwargs):
        super().__init__(**kwargs)
        if page is not None:
            self._data.page = page

    def __await__(self):
        future = aio.Future()
        self.__do__(future)
        return iter(future)

    def __do__(self, future=None):
        self._data.page.print(J()["$$BEAR"].cb(self.thunk(), future))


async def suppress_cancel(coro):
    try:
        await coro
    except aio.CancelledError:
        logger.info(f"Cancelled: {coro}")


class Page:
    def __init__(
        self,
        instance,
        selector=None,
        track_history=True,
        hgen=None,
        debug=False,
        loop=None,
        tasks=None,
    ):
        self.instance = instance
        self.iq = instance.iq
        self.oq = instance.oq
        self.query_params = instance.query_params
        self.session = instance.session
        self.representer = instance.representer
        self.hgen = hgen or StarbearHTMLGenerator(instance.representer)
        self.selector = selector
        self.track_history = track_history
        self.tasks = set() if tasks is None else tasks
        self.debug = debug
        self.loop = loop or aio.get_running_loop()
        self.js = AwaitableJ(page=self, object=self.selector)
        self.window = AwaitableJ(page=self)
        self.bearlib = AwaitableJ(page=self)["$$BEAR"]

    def __getitem__(self, selector):
        if not isinstance(selector, tuple):
            selector = (selector,)
        selector = " ".join(map(selector_for, selector))

        if self.selector is not None:
            selector = f"{self.selector} {selector}"

        return self.page_select(selector)

    def page_select(self, selector):
        return type(self)(
            instance=self.instance,
            selector=selector,
            track_history=self.track_history,
            hgen=self.hgen,
            debug=self.debug,
            loop=self.loop,
            tasks=self.tasks,
        )

    def with_history(self, track_history=True):
        if self.track_history == track_history:
            return self
        else:
            return type(self)(
                instance=self.instance,
                selector=self.selector,
                track_history=track_history,
                hgen=self.hgen,
                debug=self.debug,
                loop=self.loop,
            )

    def without_history(self):
        return self.with_history(False)

    def _done_cb(self, future):
        self.tasks.discard(future)
        if exc := future.exception():
            self.error(
                message="An error occurred trying to represent data.",
                exception=exc,
            )

    def _push(self, coro, label=None):
        if aio._get_running_loop() is None:
            aio._set_running_loop(self.loop)
        task = aio.create_task(suppress_cancel(coro), name=label)
        self.tasks.add(task)
        task.add_done_callback(self._done_cb)

    async def sync(self):
        while self.tasks:
            task = self.tasks.pop()
            await task

    def _to_element(self, x):
        if isinstance(x, str):
            return H.span(x)
        else:
            return self.hgen.hrepr(x)

    def _generate_put_commands(self, element, method, send_resources=False):
        sel = self.selector or "body"
        if not element:
            yield {
                "command": "put",
                "selector": sel,
                "method": method,
                "content": "",
            }
            return

        blk = self.hgen.blockgen(element)

        if send_resources and blk.processed_resources:
            yield {
                "command": "resource",
                "content": str(H.inline(blk.processed_resources)),
            }
        yield {
            "command": "put",
            "selector": sel,
            "method": method,
            "content": str(blk.result),
        }
        for xtra in blk.processed_extra:
            s = str(xtra.start)
            e = str(xtra.end)
            if (
                isinstance(xtra, Breakable)
                and s.startswith("<script")
                and "src=" not in s
                and e == "</script>"
            ):
                yield {
                    "command": "eval",
                    "code": str(Sequence(*xtra.body)),
                    "module": 'type="module"' in s,
                }
            else:
                yield {
                    "command": "put",
                    "selector": sel,
                    "method": "beforeend",
                    "content": str(H.div(xtra, style="display:none")),
                }
        for elem_id, (lg, listeners) in blk.live_generators.items():
            coro = lg(self[f"#{elem_id}"])
            if inspect.isasyncgen(coro):
                self._push(self._wrap_async_gen(coro, listeners), label=elem_id)
            else:
                self._push(coro, label=elem_id)

    async def _wrap_async_gen(self, agen, listeners):
        agen = aiter(agen)
        send_back = None

        while True:
            try:
                result = await agen.asend(send_back)
                processed = False
                keys = [True]
                if isinstance(result, Event):
                    keys.append(result.type)
                for key in keys:
                    method = listeners.get(key, None)
                    if isinstance(method, Queue):
                        method.put_nowait(result)
                        processed = True
                    elif method is None:
                        pass
                    else:
                        method(result)
                        processed = True
                if isinstance(result, FeedbackEvent):
                    if not processed:
                        result.resolve(Responses.NO_LISTENERS)
                    send_back = result.response
                else:
                    send_back = None

            except StopAsyncIteration:
                break

            except Exception as exc:
                self.error(f"Error in {agen}", exception=exc)
                raise

    async def put(self, element, method, history=None, send_resources=True):
        if history is None:
            history = self.track_history
        await self.oq.put(
            (
                list(self._generate_put_commands(element, method, send_resources)),
                history,
            )
        )

    def put_nowait(self, element, method, history=None, send_resources=True):
        self._push(self.put(element, method, history=history, send_resources=send_resources))

    def queue_command(self, command, /, history=None, **arguments):
        if history is None:
            history = self.track_history
        arguments["command"] = command
        self._push(self.oq.put((arguments, history)))

    def set_title(self, title):
        self.queue_command("put", selector="head title", content=title, method="innerHTML")

    def add_resources(self, *resources, type=None):
        def _build(resource, name):
            if name.endswith(".css") or type == "text/css":
                return H.link(rel="stylesheet", href=resource)
            elif name.endswith(".js") or type == "text/javascript":
                return H.script(src=resource)
            elif name.endswith(".ico"):
                return H.link(rel="icon", href=resource)
            else:
                raise ValueError(f"Cannot determine resource type for '{resource}'")

        for resource in resources:
            if isinstance(resource, str):
                if resource.startswith("http://") or resource.startswith("https://"):
                    node = _build(resource, resource)
                else:
                    node = _build(Path(resource), resource)
            elif isinstance(resource, Path):
                node = _build(resource, resource.suffix)
            elif isinstance(resource, Tag):
                node = resource
            else:
                raise TypeError("resource argument should be a Path or a Tag object")

            text = self.hgen.to_string(node)
            self.queue_command("resource", content=text)

    def print(self, *elements, method="beforeend"):
        for element in elements:
            element = self._to_element(element)
            self.put_nowait(element, method)

    def error(self, message, debug=None, exception=None):
        if not isinstance(message, str):
            message = str(self.hgen.hrepr(message))
        self.queue_command("error", content=format_error(message, debug, exception, self.debug))

    def log(self, message):
        if not isinstance(message, str):
            message = self.representer.hrepr(message)
        self.queue_command("log", content=str(message))

    def print_html(self, html, selector=None, method="beforeend", history=None):
        self.queue_command(
            "put",
            selector=selector or self.selector or "body",
            method=method,
            content=html,
            history=history,
        )

    def template(self, template_file, integration_method="innerHTML", **params):
        filled = self.instance.template(template_file, **params)
        self.put_nowait(filled, integration_method)

    def set(self, element):
        element = self._to_element(element)
        self.put_nowait(element, "innerHTML")

    def replace(self, element):
        element = self._to_element(element)
        self.put_nowait(element, "outerHTML")

    def clear(self):
        self.put_nowait("", "innerHTML")

    def delete(self):
        self.put_nowait("", "outerHTML")

    async def eval(self, code):
        return await self.js.eval(code)

    def exec(self, code, future=None):
        self.js.exec(code).__do__(future)

    def toggle(self, toggle, value=None):
        return self.bearlib.toggle(self, toggle, value)

    def __hrepr__(self, H, hrepr):
        return hrepr.make.instance(
            title="Page",
            fields=[["selector", self.selector]],
            delimiter=":",
        )

    def __js_embed__(self, representer):
        if self.selector is None:
            raise Exception("Cannot send page reference to JS because it has no selector.")
        return f"document.querySelector('{self.selector}')"

    async def recv(self):
        return await self.iq.get()

    async def wait(self, timeout=None):
        await aio.sleep(1_000_000_000 if timeout is None else timeout)

    @property
    def server_instance(self):
        return self.instance.mother.app.starbear_instance



import asyncio
import atexit
import json
import logging
import os
import sys
from asyncio import Future
from pathlib import Path
from uuid import uuid4

from sse_starlette.sse import EventSourceResponse
from starlette.routing import Route
from watchdog.observers import Observer

logger = logging.getLogger("starbear")


def restart():  # pragma: no cover
    # Note: It may be difficult to get coverage for this file, because restart() may prevent coverage
    # data from being written to the file.

    def x(args):
        atexit._run_exitfuncs()
        os.execl(sys.executable, sys.executable, *args)  # pragma: no cover

    p = Path(sys.argv[0])
    if p.name == "__main__.py":
        init = p.parent / "__init__.py"
        for m, obj in sys.modules.items():
            if getattr(obj, "__file__", None) == str(init):
                x(["-m", m, *sys.argv[1:]])
        else:
            raise SystemError("Cannot figure out how to restart starbear.")
    else:
        x(sys.argv)


class Looper:
    def __init__(self, reloader):
        self.future = None
        self.reloader = reloader
        self.reloader.activity.append(self.handle)

    def handle(self):
        if self.future and not self.future.done():
            self.future.set_result(True)

    async def __aiter__(self):
        try:
            while True:
                self.future = Future()
                await self.future
                await asyncio.sleep(0.05)
                yield True
        except asyncio.CancelledError:
            self.reloader.activity.remove(self.handle)


async def autofire():
    yield True


class InertReloader:
    def __init__(self, server):
        self.server = server
        self.uuid = uuid4().hex
        self.activity = []
        self.has_fired = False

    def prep(self):
        pass

    def code_watch(self, watch):
        pass

    def browser_side_code(self):
        return None

    def inject_routes(self, routes):
        pass


class BaseReloader(InertReloader):
    def browser_side_code(self):
        return f"""
        let src = new EventSource("/!!{self.uuid}/events");
        function reboot() {{
            fetch("/!!{self.uuid}/reboot");
            setTimeout(() => window.location.reload(), 500);
        }}
        $$BEAR.tabs.addButton("⟳", reboot);
        """

    def code_watch(self, watch):
        self.obs = Observer()
        if not isinstance(watch, list):  # pragma: no cover
            watch = [watch]
        for w in watch:
            self.obs.schedule(self, w, recursive=True)
        self.obs.start()

    async def route_events(self, request):
        return EventSourceResponse(Looper(self))

    async def reboot(self, request):
        logger.info("Rebooting the server...")
        os.environ["STARBEAR_RELOAD_OVERRIDE"] = json.dumps(
            tuple(self.server.config.socket.getsockname())
        )
        self.server.config.socket.close()
        restart()

    def inject_routes(self, routes):
        routes.insert(0, Route(f"/!!{self.uuid}/events", self.route_events))
        routes.insert(0, Route(f"/!!{self.uuid}/reboot", self.reboot))

    def is_watched(self, pth):
        return not ("__pycache__" in pth or pth.endswith(".pyc"))

    def fire(self):
        self.has_fired = True
        for listener in self.activity:
            listener()

    def dispatch(self, event):
        if self.is_watched(event.src_path):
            self.fire()


class JuriggedReloader(BaseReloader):
    def __init__(self, server):
        from jurigged.codetools import CodeFileOperation
        from jurigged.register import registry

        super().__init__(server)
        self.registry = registry
        self.op = CodeFileOperation

    def prep(self):
        # Sometimes has to be done before importing the module to watch in order
        # to properly collect function data
        import codefind  # noqa: F401

    def code_watch(self, watch):
        import jurigged

        super().code_watch(watch)
        jurigged.watch(watch)
        self.activity.append(self.server.inject_routes)
        self.registry.activity.register(self.handle_jurigged)

    def browser_side_code(self):
        code = super().browser_side_code()
        return (
            code
            + """
        src.onmessage = e => {
            window.location.reload();
        };
        """
        )

    def handle_jurigged(self, event):
        if isinstance(event, self.op):
            self.fire()

    def is_watched(self, pth):
        return super().is_watched(pth) and not pth.endswith(".py")


class InertJuriggedReloader(JuriggedReloader):
    def browser_side_code(self):
        return ""

    def handle_jurigged(self, event):
        pass


class FullReloader(BaseReloader):
    def __init__(self, server):
        super().__init__(server)
        self.donezo = False

    async def route_events(self, request):
        if self.has_fired:
            return EventSourceResponse(autofire())
        else:
            return EventSourceResponse(Looper(self))

    def browser_side_code(self):
        code = super().browser_side_code()
        return (
            code
            + """
        src.onmessage = reboot;
        """
        )



import time

from hrepr import H

from starbear import FeedbackQueue, Resource as R, bear


@bear
async def __app__(page):
    q = FeedbackQueue()

    def f(x):
        page[out].set(H.div(f"Current square: {x}"))

    page.print(
        H.script(f"run = async e => {R(f)}((await {R(q)}(e)) ** 2)"),
        H.div(H.button("Clicky!", onclick="run(event)")),
        out := H.div(id="result"),
    )
    current = 0
    async for event, resolve in q:
        await resolve(current)
        current += 1


def test_feedback(app):
    result = app.locator("#result")
    app.locator("button").click()
    time.sleep(0.1)
    assert result.inner_text() == "Current square: 0"
    app.locator("button").click()
    time.sleep(0.1)
    assert result.inner_text() == "Current square: 1"
    app.locator("button").click()
    time.sleep(0.1)
    assert result.inner_text() == "Current square: 4"



import threading
import time
import webbrowser
from contextlib import contextmanager
from functools import cached_property, partial
from graphlib import TopologicalSorter
from pathlib import Path
from textwrap import dedent

import gifnoc
import uvicorn
from hrepr import H
from starlette.applications import Starlette
from starlette.middleware.httpsredirect import HTTPSRedirectMiddleware

from ..common import logger
from ..config import config as base_config
from .config import StarbearServerConfig
from .find import compile_routes
from .plugins.session import Session


class ThreadableServer(uvicorn.Server):
    def install_signal_handlers(self):
        pass

    def run(self, config=None):
        with gifnoc.use(config or {}):
            super().run()

    @contextmanager
    def run_in_thread(self, config=None):
        # Code taken from https://stackoverflow.com/questions/61577643/python-how-to-use-fastapi-and-uvicorn-run-without-blocking-the-thread
        thread = threading.Thread(target=partial(self.run, config))
        thread.start()
        try:
            while not self.started:
                time.sleep(1e-3)
            yield
        finally:
            self.should_exit = True
            thread.join()


class StarbearServer:
    def __init__(self, config: StarbearServerConfig):
        self.config = config

    @cached_property
    def reloader(self):
        return self.config.get_reloader(self)

    def get_routes(self):
        return self.config.get_routes()

    def get_locations(self):
        return self.config.get_locations()

    def inject_routes(self):
        collected = self.get_routes()
        routes = compile_routes("/", collected)
        self.reloader.inject_routes(routes)

        for route in routes:
            route._starbear_managed = True

        remainder = [
            r for r in self.app.router.routes if not getattr(r, "_starbear_managed", False)
        ]
        self.app.router.routes = remainder + routes
        self.app.map = collected

    def _setup(self):
        if self.config.watch is True:
            watch = self.get_locations()
        else:
            watch = self.config.watch

        self.reloader.prep()
        self.reloader.code_watch(watch)

        inject_code = self.reloader.browser_side_code()
        if inject_code:
            self.inject = [
                H.script(
                    dedent(
                        f"""window.addEventListener("load", () => {{
                        {inject_code}
                        }});
                        """
                    )
                )
            ]
        else:
            self.inject = []

        app = Starlette(routes=[])

        @app.on_event("startup")
        async def _():
            protocol = "https" if self.config.ssl.enabled else "http"
            host, port = self.config.socket.getsockname()
            url = f"{protocol}://{host}:{port}"
            logger.info(f"Serving at: {url}")
            if self.config.open_browser:
                webbrowser.open(url)

        def _ensure(filename, enabled):
            if not enabled or not filename:
                return None
            if not Path(filename).exists():
                raise FileNotFoundError(filename)
            return filename

        ssl_enabled = self.config.ssl.enabled
        self.ssl_keyfile = _ensure(self.config.ssl.keyfile, ssl_enabled)
        self.ssl_certfile = _ensure(self.config.ssl.certfile, ssl_enabled)

        if ssl_enabled and self.ssl_certfile and self.ssl_keyfile:
            # This doesn't seem to do anything?
            app.add_middleware(HTTPSRedirectMiddleware)

        self.app = app

        plugins = {name: p for name, p in self.config.plugins.items() if p.enabled}

        exports = {}
        for i, p in plugins.items():
            exports.update({x: i for x in p.cap_export()})

        if "session" not in exports:
            assert "session" not in plugins
            exports["session"] = "session"
            plugins["session"] = Session(enabled=True, required=False)

        def locate_dependencies(p):
            for x in p.cap_require():
                if x not in exports:
                    raise Exception(
                        f"Plugin '{i}' of type '{type(p).__name__}' requires the '{x}' feature, but none of the included plugins exports it."
                    )
                plugins[exports[x]].required = True
                yield exports[x]

        graph = {i: list(locate_dependencies(p)) for i, p in plugins.items()}
        order = TopologicalSorter(graph).static_order()

        for plugin_index in reversed(list(order)):
            p = plugins[plugin_index]
            if p.required:
                assert not hasattr(self, plugin_index)
                setattr(self, plugin_index, p)
                logger.info(f"Set up plugin: {plugin_index}")
                p.setup(self)

        self.plugins = plugins
        self.app.starbear_instance = self

        self.inject_routes()

    @contextmanager
    def _server(self, *, thread=False, **uvicorn_options):
        self._setup()
        server_class = ThreadableServer if thread else uvicorn.Server
        with gifnoc.overlay(
            {
                "starbear": {
                    "dev": {
                        "debug_mode": base_config.dev.debug_mode or self.config.dev,
                        "inject": self.inject,
                    }
                }
            }
        ):
            uconfig = uvicorn.Config(
                app=self.app,
                fd=self.config.socket.fileno(),
                log_level="info",
                ssl_keyfile=self.ssl_keyfile,
                ssl_certfile=self.ssl_certfile,
                **uvicorn_options,
            )
            yield server_class(uconfig)

    def run(self, **uvicorn_options):
        with self._server(thread=False, **uvicorn_options) as server:
            server.run()

    @contextmanager
    def run_in_thread(self, **uvicorn_options):
        with self._server(thread=True, **uvicorn_options) as server:
            with server.run_in_thread():
                yield


def run(**kwargs):
    server = StarbearServer(StarbearServerConfig(**kwargs))
    server.run()



from hrepr import H, J

from starbear import simplebear

from .utils import asset_getter

asset = asset_getter(__file__)


async def exponential(x):
    return (x * 2) or 1


@simplebear
async def __app__(request):
    return H.div(
        H.h1("Hello!", id="content"),
        J(module=asset("counter.js"))(increment=10, cls="c1"),
        J(module=asset("complex-counter.js"))(increment=exponential, cls="c2"),
    )


def test_simple(app):
    assert app.locator("#content").inner_text() == "Hello!"


def test_counter(app):
    counter = app.locator(".c1")
    value0 = int(counter.inner_text())
    counter.click()
    value1 = int(counter.inner_text())
    counter.click()
    value2 = int(counter.inner_text())
    assert value2 == value1 + 10 == value0 + 20


def test_complex_counter(app):
    counter = app.locator(".c2")
    for i in range(5):
        counter.click()
        assert int(counter.inner_text()) == 2**i



from pathlib import Path
from shutil import copytree

import pytest


@pytest.fixture
def clone(tmpdir):
    def _clone(pth):
        dest = tmpdir / pth.name
        copytree(src=pth, dst=dest)
        return Path(str(dest))

    return _clone



import time
from contextlib import contextmanager

import pytest

from starbear import here

from .utils import serve


@pytest.fixture
def venus_mod(page, clone):
    tmp = clone(here / "app_hello")

    @contextmanager
    def load(*args, regoto=False):
        page.set_default_timeout(2500)
        with serve(tmp, *args) as addr:
            page.goto(f"{addr}/world/venus")
            h1 = page.locator("h1")
            assert h1.inner_text() == "VENUS!"
            vfile = tmp / "world" / "venus.py"
            vfile.write_text(vfile.read_text().replace("VENUS!", "VENUSAUR!"))
            yield page
            if regoto:
                page.goto(f"{addr}/world/venus")
            assert h1.inner_text() == "VENUSAUR!"

    return load


@pytest.fixture
def pause(request):
    return request.config.getoption("--reload-pause")


def test_manual(venus_mod, pause):
    with venus_mod("--dev", "--reload-mode", "manual", regoto=True) as page:
        page.locator('.bear--tabular-button:has-text("⟳")').click()
        time.sleep(pause)


def test_jurigged(venus_mod, pause):
    with venus_mod("--dev"):
        time.sleep(pause)


def test_full(venus_mod, pause):
    with venus_mod("--dev", "--reload-mode", "full", regoto=True):
        time.sleep(pause)
